#include <iostream>
#include <string>
#include <vector>
#include <stdexcept>
#include <stdio.h>
#include <cstdio>
#include <memory>
#include <array>
#include <chrono>
#include <thread>
#include <fstream>

#include "tgvoip/VoIPController.h"
#include "tgvoip/VoIPServerConfig.h"
#include "tgvoip/audio/AudioIOCallback.h"

using namespace std;
using namespace tgvoip;

const string CALLER = "caller";
const string CALLEE = "callee";

vector<int16_t> exec(const char* cmd) {
  std::array<int16_t, 128> buffer;
  vector<int16_t> result;
  std::unique_ptr<FILE, decltype(&pclose)> pipe(popen(cmd, "r"), pclose);
  if (!pipe) {
    throw std::runtime_error("popen() failed!");
  }
  while (fread(buffer.data(), sizeof(int16_t), buffer.size(), pipe.get()) !=
         0) {
    result.insert(result.end(), buffer.begin(), buffer.end());
  }
  return result;
}

int to_hex(char a) {
  if (a >= '0' && a <= '9') return (a - '0');
  return 10 + (a - 'a');
}

class VoIPControllerWrapper {
 private:
  std::unique_ptr<VoIPController> cnt;
  FILE* f_in;
  FILE* f_out;
  Mutex m_in;
  Mutex m_out;
  bool sending_finished = false;
  vector<int16_t> in_content;
  int in_offset = 0;
  std::chrono::time_point<std::chrono::system_clock> sending_over;

 public:
  VoIPControllerWrapper(char* key, int port, string addr, char* tag,
                        bool is_outgoing, long long id = 1) {
    cnt = std::make_unique<VoIPController>();
    VoIPController::Config config;
    config.enableNS = config.enableAEC = config.enableAGC = false;
    config.enableCallUpgrade = false;
    config.recvTimeout = config.initTimeout = 5.0;
    cnt->SetConfig(config);

    vector<Endpoint> endpoints;
    IPv4Address v4(addr);
    IPv6Address v6("::0");  // not in scope for the contest
    auto* peer_tag = new unsigned char[16];
    for (int i = 0; i < 16; i++) {
      peer_tag[i] = to_hex(tag[2 * i]) * 16 + to_hex(tag[2 * i + 1]);
    }
    endpoints.push_back(
        Endpoint(id, port, v4, v6, Endpoint::Type::TCP_RELAY, peer_tag));
    endpoints.push_back(
        Endpoint(id, port, v4, v6, Endpoint::Type::UDP_RELAY, peer_tag));

    unsigned char ekey[256];
    for (int i = 0; i < 256; i++) {
      ekey[i] = to_hex(key[i * 2]) * 16 + to_hex(key[i * 2 + 1]);
    }
    cnt->SetEncryptionKey(reinterpret_cast<char*>(ekey),
                          /*is_outgoing*/ is_outgoing);
    cnt->SetRemoteEndpoints(endpoints, false,
                            VoIPController::GetConnectionMaxLayer());
  }

  void setAudioDataCallbacks(string p_in, string p_out) {
    string cmd = "ffmpeg -loglevel panic -i " + p_in +
                 " -f s16le -ac 1 -ar 48000 -acodec pcm_s16le -";
    this->in_content = exec(cmd.c_str());

    f_in = fopen(p_in.c_str(), "rb");

    string cmd2 =
        "ffmpeg -loglevel panic -f s16le -ac 1 -ar 48000 -acodec "
        "pcm_s16le -i - -b:a 64000 " +
        p_out;
    f_out = popen(cmd2.c_str(), "w");

    cnt->SetAudioDataCallbacks(
        [this](int16_t* data, size_t len) {
          // Send audio
          MutexGuard mg(this->m_in);
          if (this->f_in == NULL) {
            memset(data, 0, len);
            return;
          }
          size_t readInput =
              min(len, this->in_content.size() - this->in_offset);
          std::copy(this->in_content.begin() + in_offset,
                    this->in_content.begin() + this->in_offset + readInput,
                    data);
          if (readInput != len) {
            fclose(this->f_in);
            this->f_in = NULL;
            memset(data + (readInput % len), 0, len - (readInput % len));
            this->sending_finished = true;
            this->sending_over = std::chrono::system_clock::now();
          }
          in_offset += len;

        },
        [this](int16_t* data, size_t len) {
          // Receive audio
          MutexGuard mg(this->m_out);
          if (this->f_out != NULL) {
            size_t actual_len = 0;
            while (actual_len < len && data[actual_len] != 0) {
              actual_len++;
            }
            len = actual_len;

            if (fwrite(data, sizeof(int16_t), len, this->f_out) != len) {
              // cerr << "Error: could not write the data";
            }
          }

        });
  }

  ~VoIPControllerWrapper() {
    cout << cnt->GetDebugLog() << endl;
    cnt->Stop();
    if (f_in) {
      fclose(f_in);
      f_in = NULL;
    }
    if (f_out) {
      fclose(f_out);
      f_out = NULL;
    }
  }

  int run() {
    cnt->Start();
    cnt->Connect();
    while (1) {
      std::this_thread::sleep_for(std::chrono::milliseconds(500));
      MutexGuard mg(m_in);
      if (sending_finished) {
        auto now = std::chrono::system_clock::now();
        auto time_left = std::chrono::duration_cast<std::chrono::milliseconds>(
            now - this->sending_over);
        std::this_thread::sleep_for(std::chrono::milliseconds(5 * 1000) -
                                    time_left);
        return 0;
      }
      if (cnt->GetConnectionState() == STATE_FAILED) {
        return 1;
      }
    }
  }
};

int main(int argc, char* argv[]) {
  string host = argv[1];
  auto e = host.find(":");
  string addr = host.substr(0, e);
  int port = atoi(host.substr(e + 1).c_str());

  char* tag = argv[2];
  char* encryptionkey = argv[4];
  string input = argv[6];
  string output = argv[8];
  string config = argv[10];
  string type = argv[12];

  std::ifstream ifs(config.c_str());
  std::string config_str((std::istreambuf_iterator<char>(ifs)),
                         (std::istreambuf_iterator<char>()));

  ServerConfig::GetSharedInstance()->Update(config_str);

  VoIPControllerWrapper cnt(encryptionkey, port, addr, tag, type == CALLER,
                            1ll);
  cnt.setAudioDataCallbacks(input, output);
  return cnt.run();
}
